"""Defines the templaters."""

import ast
import re
from collections.abc import Iterable, Iterator, Mapping
from string import Formatter
from typing import Any, Callable, NamedTuple, Optional

from sqlfluff.core.config import FluffConfig
from sqlfluff.core.errors import SQLTemplaterError
from sqlfluff.core.formatter import FormatterInterface
from sqlfluff.core.helpers.slice import offset_slice, zero_slice
from sqlfluff.core.helpers.string import findall
from sqlfluff.core.templaters.base import (
    RawFileSlice,
    RawTemplater,
    TemplatedFile,
    TemplatedFileSlice,
    large_file_check,
    templater_logger,
)


class IntermediateFileSlice(NamedTuple):
    """An intermediate representation of a partially sliced File."""

    intermediate_type: str
    source_slice: slice
    templated_slice: slice
    slice_buffer: list[RawFileSlice]

    def _trim_end(
        self, templated_str: str, target_end: str = "head"
    ) -> tuple["IntermediateFileSlice", list[TemplatedFileSlice]]:
        """Trim the ends of a intermediate segment."""
        target_idx = 0 if target_end == "head" else -1
        terminator_types = ("block_start") if target_end == "head" else ("block_end")
        main_source_slice = self.source_slice
        main_templated_slice = self.templated_slice
        slice_buffer = self.slice_buffer

        end_buffer = []

        # Yield any leading literals, comments or blocks.
        while len(slice_buffer) > 0 and slice_buffer[target_idx].slice_type in (
            "literal",
            "block_start",
            "block_end",
            "comment",
        ):
            focus = slice_buffer[target_idx]
            templater_logger.debug("            %s Focus: %s", target_end, focus)
            # Is it a zero length item?
            if focus.slice_type in ("block_start", "block_end", "comment"):
                # Only add the length in the source space.
                templated_len = 0
            else:
                # Assume it's a literal, check the literal actually matches.
                templated_len = len(focus.raw)
                if target_end == "head":
                    check_slice = offset_slice(
                        main_templated_slice.start,
                        templated_len,
                    )
                else:
                    check_slice = slice(
                        main_templated_slice.stop - templated_len,
                        main_templated_slice.stop,
                    )

                if templated_str[check_slice] != focus.raw:
                    # It doesn't match, we can't use it. break
                    templater_logger.debug("                Nope")
                    break

            # If it does match, set up the new slices
            if target_end == "head":
                division = (
                    main_source_slice.start + len(focus.raw),
                    main_templated_slice.start + templated_len,
                )
                new_slice = TemplatedFileSlice(
                    focus.slice_type,
                    slice(main_source_slice.start, division[0]),
                    slice(main_templated_slice.start, division[1]),
                )
                end_buffer.append(new_slice)
                main_source_slice = slice(division[0], main_source_slice.stop)
                main_templated_slice = slice(division[1], main_templated_slice.stop)
            else:
                division = (
                    main_source_slice.stop - len(focus.raw),
                    main_templated_slice.stop - templated_len,
                )
                new_slice = TemplatedFileSlice(
                    focus.slice_type,
                    slice(division[0], main_source_slice.stop),
                    slice(division[1], main_templated_slice.stop),
                )
                end_buffer.insert(0, new_slice)
                main_source_slice = slice(main_source_slice.start, division[0])
                main_templated_slice = slice(main_templated_slice.start, division[1])

            slice_buffer.pop(target_idx)
            if focus.slice_type in terminator_types:
                break
        # Return a new Intermediate slice and the buffer.
        # NB: Don't check size of slice buffer here. We can do that later.
        new_intermediate = self.__class__(
            "compound", main_source_slice, main_templated_slice, slice_buffer
        )
        return new_intermediate, end_buffer

    def trim_ends(
        self, templated_str: str
    ) -> tuple[
        list[TemplatedFileSlice], "IntermediateFileSlice", list[TemplatedFileSlice]
    ]:
        """Trim both ends of an intermediate slice."""
        # Trim start:
        new_slice, head_buffer = self._trim_end(
            templated_str=templated_str, target_end="head"
        )
        # Trim end:
        new_slice, tail_buffer = new_slice._trim_end(
            templated_str=templated_str, target_end="tail"
        )
        # Return
        return head_buffer, new_slice, tail_buffer

    def try_simple(self) -> TemplatedFileSlice:
        """Try to turn this intermediate slice into a simple slice."""
        # Yield anything simple
        if len(self.slice_buffer) == 1:
            return TemplatedFileSlice(
                self.slice_buffer[0].slice_type,
                self.source_slice,
                self.templated_slice,
            )
        else:
            raise ValueError("IntermediateFileSlice is not simple!")

    def coalesce(self) -> TemplatedFileSlice:
        """Coalesce this whole slice into a single one. Brutally."""
        return TemplatedFileSlice(
            PythonTemplater._coalesce_types(self.slice_buffer),
            self.source_slice,
            self.templated_slice,
        )


class PythonTemplater(RawTemplater):
    """A templater using python format strings.

    See: https://docs.python.org/3/library/string.html#format-string-syntax

    For the python templater we don't allow functions or macros because there isn't
    a good way of doing it securely. Use the jinja templater for this.

    The python templater also defines a lot of the logic for how
    to allow fixing and translation in a templated file.
    """

    name = "python"
    config_subsection: tuple[str, ...] = ("context",)

    def __init__(self, override_context: Optional[dict[str, Any]] = None) -> None:
        self.default_context = dict(test_value="__test__")
        self.override_context = override_context or {}

    @staticmethod
    def infer_type(s: Any) -> Any:
        """Infer a python type from a string and convert.

        Given a string value, convert it to a more specific built-in Python type
        (e.g. int, float, list, dictionary) if possible.

        """
        try:
            return ast.literal_eval(s)
        except (SyntaxError, ValueError):
            return s

    def get_context(
        self,
        fname: Optional[str],
        config: Optional[FluffConfig],
    ) -> dict[str, Any]:
        """Get the templating context from the config.

        This function retrieves the templating context from the config by
        loading the config and updating the live_context dictionary with the
        loaded_context and other predefined context dictionaries. It then goes
        through the loaded_context dictionary and infers the types of the values
        before returning the live_context dictionary.

        Args:
            fname (str, optional): The file name.
            config (dict, optional): The config dictionary.

        Returns:
            dict: The templating context.
        """
        live_context = super().get_context(fname, config)
        # Infer types
        for k in live_context:
            live_context[k] = self.infer_type(live_context[k])
        return live_context

    @large_file_check
    def process(
        self,
        *,
        in_str: str,
        fname: str,
        config: Optional[FluffConfig] = None,
        formatter: Optional[FormatterInterface] = None,
    ) -> tuple[TemplatedFile, list[SQLTemplaterError]]:
        """Process a string and return a TemplatedFile.

        Note that the arguments are enforced as keywords
        because Templaters can have differences in their
        `process` method signature.
        A Templater that only supports reading from a file
        would need the following signature:
            process(*, fname, in_str=None, config=None)
        (arguments are swapped)

        Args:
            in_str (:obj:`str`): The input string.
            fname (:obj:`str`, optional): The filename of this string. This is
                mostly for loading config files at runtime.
            config (:obj:`FluffConfig`): A specific config to use for this
                templating operation. Only necessary for some templaters.
            formatter (:obj:`CallbackFormatter`): Optional object for output.

        """
        live_context = self.get_context(fname, config)
        ignore_templating = config and "templating" in config.get("ignore")

        def render_func(raw_str: str) -> str:
            """Render the string using the captured live_context.

            In order to support mocking of template variables
            containing "." characters, this function converts any
            template variable containing "." into a dictionary lookup.
                Example:  {foo.bar} => {sqlfluff[foo.bar]}
            """
            # Hack to allow template variables with dot notation (e.g. foo.bar)
            raw_str_with_dot_notation_hack = re.sub(
                r"{([^:}]*\.[^:}]*)(:\S*)?}", r"{sqlfluff[\1]\2}", raw_str
            )
            templater_logger.debug(
                "    Raw String with Dot Notation Hack: %r",
                raw_str_with_dot_notation_hack,
            )

            if ignore_templating:
                # When ignoring templating errors, use a fallback dict that
                # returns the variable name for missing keys, so that templating
                # can still succeed and linting can proceed on the result.
                # This mirrors the behaviour of Jinja's DummyUndefined when
                # ignore=templating is set. No violations are recorded for
                # missing variables since the user has explicitly opted to
                # ignore templating errors.

                class _FallbackValue:
                    """A fallback value that acts as a placeholder for missing vars.

                    Supports item access (for dot notation variables) by returning
                    the key with dots replaced by underscores.
                    """

                    def __init__(self, name: str) -> None:
                        self.name = name

                    def __str__(self) -> str:
                        return self.name

                    def __format__(self, format_spec: str) -> str:
                        return str(self)

                    def __getitem__(self, key: str) -> str:
                        # Handles the dot-notation hack: {sqlfluff[foo.bar]}
                        return key.replace(".", "_")

                class _FallbackDict(dict[str, Any]):
                    @classmethod
                    def from_mapping(
                        cls, mapping: Mapping[str, Any]
                    ) -> "_FallbackDict":
                        """Recursively wrap mappings so nested lookups also fallback."""
                        return cls(
                            {
                                key: cls.from_mapping(value)
                                if isinstance(value, Mapping)
                                else value
                                for key, value in mapping.items()
                            }
                        )

                    def __missing__(self, key: str) -> "_FallbackValue":
                        return _FallbackValue(key)

                class _DotNotationFallbackDict(_FallbackDict):
                    def __missing__(self, key: str) -> "_FallbackValue":
                        return _FallbackValue(key.replace(".", "_"))

                fallback_context = _FallbackDict.from_mapping(live_context)
                if isinstance(fallback_context.get("sqlfluff"), Mapping):
                    fallback_context["sqlfluff"] = (
                        _DotNotationFallbackDict.from_mapping(
                            fallback_context["sqlfluff"]
                        )
                    )

                return raw_str_with_dot_notation_hack.format_map(fallback_context)

            try:
                rendered_str = raw_str_with_dot_notation_hack.format(**live_context)
            except KeyError as err:
                missing_key = err.args[0]
                if missing_key == "sqlfluff":
                    # Give more useful error message related to dot notation hack
                    # when user has not created the required, magic context key
                    raise SQLTemplaterError(
                        "Failure in Python templating: magic key 'sqlfluff' "
                        "missing from context.  This key is required "
                        "for template variables containing '.'. "
                        "https://docs.sqlfluff.com/en/stable/"
                        "perma/python_templating.html"
                    )
                elif "." in missing_key:
                    # Give more useful error message related to dot notation hack
                    # for missing keys
                    raise SQLTemplaterError(
                        "Failure in Python templating: {} key missing from 'sqlfluff' "
                        "dict in context. Template variables containing '.' are "
                        "required to use the 'sqlfluff' magic fixed context key. "
                        "https://docs.sqlfluff.com/en/stable/"
                        "perma/python_templating.html".format(err)
                    )
                else:
                    raise SQLTemplaterError(
                        "Failure in Python templating: {}. Have you configured your "
                        "variables? https://docs.sqlfluff.com/en/stable/"
                        "perma/variables.html".format(err)
                    )
            return rendered_str

        raw_sliced, sliced_file, new_str = self.slice_file(
            in_str,
            render_func=render_func,
            config=config,
        )
        return (
            TemplatedFile(
                source_str=in_str,
                templated_str=new_str,
                fname=fname,
                sliced_file=sliced_file,
                raw_sliced=raw_sliced,
            ),
            [],
        )

    def slice_file(
        self,
        raw_str: str,
        render_func: Callable[[str], str],
        config: Optional[FluffConfig] = None,
        append_to_templated: str = "",
    ) -> tuple[list[RawFileSlice], list[TemplatedFileSlice], str]:
        """Slice the file to determine regions where we can fix."""
        templater_logger.info("Slicing File Template")
        templater_logger.debug("    Raw String: %r", raw_str)
        # Render the templated string.
        # NOTE: This seems excessive in this simple example, but for other templating
        # engines we need more control over the rendering so may need to call this
        # method more than once.
        templated_str = render_func(raw_str)
        templater_logger.debug("    Templated String: %r", templated_str)
        # Slice the raw file
        raw_sliced = list(self._slice_template(raw_str))
        templater_logger.debug("    Raw Sliced:")
        for idx, raw_slice in enumerate(raw_sliced):
            templater_logger.debug("        %s: %r", idx, raw_slice)
        # Find the literals
        literals = [
            raw_slice.raw
            for raw_slice in raw_sliced
            if raw_slice.slice_type == "literal"
        ]
        templater_logger.debug("    Literals: %s", literals)
        for loop_idx in range(2):
            templater_logger.debug("    # Slice Loop %s", loop_idx)
            # Calculate occurrences
            raw_occurrences = self._substring_occurrences(raw_str, literals)
            templated_occurrences = self._substring_occurrences(templated_str, literals)
            templater_logger.debug(
                "    Occurrences: Raw: %s, Templated: %s",
                raw_occurrences,
                templated_occurrences,
            )
            # Split on invariants
            split_sliced = list(
                self._split_invariants(
                    raw_sliced,
                    literals,
                    raw_occurrences,
                    templated_occurrences,
                    templated_str,
                )
            )
            templater_logger.debug("    Split Sliced:")
            for idx, split_slice in enumerate(split_sliced):
                templater_logger.debug("        %s: %r", idx, split_slice)
            # Deal with uniques and coalesce the rest
            sliced_file = list(
                self._split_uniques_coalesce_rest(
                    split_sliced, raw_occurrences, templated_occurrences, templated_str
                )
            )
            templater_logger.debug("    Fully Sliced:")
            for idx, templ_slice in enumerate(sliced_file):
                templater_logger.debug("        %s: %r", idx, templ_slice)
            unwrap_wrapped = (
                True
                if config is None
                else config.get(
                    "unwrap_wrapped_queries", section="templater", default=True
                )
            )
            sliced_file, new_templated_str = self._check_for_wrapped(
                sliced_file, templated_str, unwrap_wrapped=unwrap_wrapped
            )
            if new_templated_str == templated_str:
                # If we didn't change it then we're done.
                break
            else:
                # If it's not equal, loop around
                templated_str = new_templated_str
        return raw_sliced, sliced_file, new_templated_str

    @classmethod
    def _check_for_wrapped(
        cls,
        slices: list[TemplatedFileSlice],
        templated_str: str,
        unwrap_wrapped: bool = True,
    ) -> tuple[list[TemplatedFileSlice], str]:
        """Identify a wrapped query (e.g. dbt test) and handle it.

        If unwrap_wrapped is true, we trim the wrapping from the templated
        file.
        If unwrap_wrapped is false, we add a slice at start and end.
        """
        if not slices:
            # If there are no slices, return
            return slices, templated_str
        first_slice = slices[0]
        last_slice = slices[-1]

        if unwrap_wrapped:
            # If we're unwrapping, there is no need to edit the slices, but we do need
            # to trim the templated string. We should expect that the template will need
            # to be re-sliced but we should assume that the function calling this one
            # will deal with that eventuality.
            return (
                slices,
                templated_str[
                    first_slice.templated_slice.start : last_slice.templated_slice.stop
                ],
            )

        if (
            first_slice.source_slice.start == 0
            and first_slice.templated_slice.start != 0
        ):
            # This means that there is text at the start of the templated file which
            # doesn't exist in the raw file. Handle this by adding a templated slice
            # (though it's not really templated) between 0 and 0 in the raw, and 0 and
            # the current first slice start index in the templated.
            slices.insert(
                0,
                TemplatedFileSlice(
                    "templated",
                    slice(0, 0),
                    slice(0, first_slice.templated_slice.start),
                ),
            )
        if last_slice.templated_slice.stop != len(templated_str):
            # This means that there is text at the end of the templated file which
            # doesn't exist in the raw file. Handle this by adding a templated slice
            # beginning and ending at the end of the raw, and the current last slice
            # stop and file end in the templated.
            slices.append(
                TemplatedFileSlice(
                    "templated",
                    zero_slice(last_slice.source_slice.stop),
                    slice(last_slice.templated_slice.stop, len(templated_str)),
                )
            )
        return slices, templated_str

    @classmethod
    def _substring_occurrences(
        cls, in_str: str, substrings: Iterable[str]
    ) -> dict[str, list[int]]:
        """Find every occurrence of the given substrings."""
        occurrences = {}
        for substring in substrings:
            occurrences[substring] = list(findall(substring, in_str))
        return occurrences

    @staticmethod
    def _sorted_occurrence_tuples(
        occurrences: dict[str, list[int]],
    ) -> list[tuple[str, int]]:
        """Sort a dict of occurrences into a sorted list of tuples."""
        return sorted(
            ((raw, idx) for raw in occurrences.keys() for idx in occurrences[raw]),
            # Sort first by position, then by lexical (for stability)
            key=lambda x: (x[1], x[0]),
        )

    @classmethod
    def _slice_template(cls, in_str: str) -> Iterator[RawFileSlice]:
        """Slice a templated python string into token tuples.

        This uses Formatter() as per:
        https://docs.python.org/3/library/string.html#string.Formatter
        """
        fmt = Formatter()
        in_idx = 0
        for literal_text, field_name, format_spec, conversion in fmt.parse(in_str):
            if literal_text:
                escape_chars = cls._sorted_occurrence_tuples(
                    cls._substring_occurrences(literal_text, ["}", "{"])
                )
                idx = 0
                while escape_chars:
                    first_char = escape_chars.pop()
                    # Is there a literal first?
                    if first_char[1] > idx:
                        yield RawFileSlice(
                            literal_text[idx : first_char[1]], "literal", in_idx
                        )
                        in_idx += first_char[1] - idx
                    # Add the escaped
                    idx = first_char[1] + len(first_char[0])
                    # We double them here to make the raw
                    yield RawFileSlice(
                        literal_text[first_char[1] : idx] * 2, "escaped", in_idx
                    )
                    # Will always be 2 in this case.
                    # This is because ALL escape sequences in the python formatter
                    # are two characters which reduce to one.
                    in_idx += 2
                # Deal with last one (if present)
                if literal_text[idx:]:
                    yield RawFileSlice(literal_text[idx:], "literal", in_idx)
                    in_idx += len(literal_text) - idx
            # Deal with fields
            if field_name:
                constructed_token = "{{{field_name}{conv}{spec}}}".format(
                    field_name=field_name,
                    conv=f"!{conversion}" if conversion else "",
                    spec=f":{format_spec}" if format_spec else "",
                )
                yield RawFileSlice(constructed_token, "templated", in_idx)
                in_idx += len(constructed_token)

    @classmethod
    def _split_invariants(
        cls,
        raw_sliced: list[RawFileSlice],
        literals: list[str],
        raw_occurrences: dict[str, list[int]],
        templated_occurrences: dict[str, list[int]],
        templated_str: str,
    ) -> Iterator[IntermediateFileSlice]:
        """Split a sliced file on its invariant literals.

        We prioritise the _longest_ invariants first as they
        are more likely to be the anchors.
        """
        # Calculate invariants
        invariants = [
            literal
            for literal in literals
            if len(raw_occurrences[literal]) == 1
            and len(templated_occurrences[literal]) == 1
        ]
        # Work through the invariants and make sure they appear
        # in order.
        for linv in sorted(invariants, key=len, reverse=True):
            # Any invariants which have templated positions, relative
            # to source positions, which aren't in order, should be
            # ignored.

            # Is this one still relevant?
            if linv not in invariants:
                continue  # pragma: no cover

            source_pos, templ_pos = raw_occurrences[linv], templated_occurrences[linv]
            # Copy the list before iterating because we're going to edit it.
            for tinv in invariants.copy():
                if tinv != linv:
                    src_dir = source_pos > raw_occurrences[tinv]
                    tmp_dir = templ_pos > templated_occurrences[tinv]
                    # If it's not in the same direction in the source and template
                    # remove it.
                    if src_dir != tmp_dir:  # pragma: no cover
                        templater_logger.debug(
                            "          Invariant found out of order: %r", tinv
                        )
                        invariants.remove(tinv)

        # Set up some buffers
        buffer: list[RawFileSlice] = []
        idx: Optional[int] = None
        templ_idx = 0
        # Loop through
        for raw_file_slice in raw_sliced:
            if raw_file_slice.raw in invariants:
                if buffer:
                    yield IntermediateFileSlice(
                        "compound",
                        slice(idx, raw_file_slice.source_idx),
                        slice(templ_idx, templated_occurrences[raw_file_slice.raw][0]),
                        buffer,
                    )
                buffer = []
                idx = None
                yield IntermediateFileSlice(
                    "invariant",
                    offset_slice(
                        raw_file_slice.source_idx,
                        len(raw_file_slice.raw),
                    ),
                    offset_slice(
                        templated_occurrences[raw_file_slice.raw][0],
                        len(raw_file_slice.raw),
                    ),
                    [
                        RawFileSlice(
                            raw_file_slice.raw,
                            raw_file_slice.slice_type,
                            templated_occurrences[raw_file_slice.raw][0],
                        )
                    ],
                )
                templ_idx = templated_occurrences[raw_file_slice.raw][0] + len(
                    raw_file_slice.raw
                )
            else:
                buffer.append(
                    RawFileSlice(
                        raw_file_slice.raw,
                        raw_file_slice.slice_type,
                        raw_file_slice.source_idx,
                    )
                )
                if idx is None:
                    idx = raw_file_slice.source_idx
        # If we have a final buffer, yield it
        if buffer:
            yield IntermediateFileSlice(
                "compound",
                slice((idx or 0), (idx or 0) + sum(len(slc.raw) for slc in buffer)),
                slice(templ_idx, len(templated_str)),
                buffer,
            )

    @staticmethod
    def _filter_occurrences(
        file_slice: slice, occurrences: dict[str, list[int]]
    ) -> dict[str, list[int]]:
        """Filter a dict of occurrences to just those within a slice."""
        filtered = {
            key: [
                pos
                for pos in occurrences[key]
                if pos >= file_slice.start and pos < file_slice.stop
            ]
            for key in occurrences.keys()
        }
        return {key: filtered[key] for key in filtered.keys() if filtered[key]}

    @staticmethod
    def _coalesce_types(elems: list[RawFileSlice]) -> str:
        """Coalesce to the priority type."""
        # Make a set of types
        types = {elem.slice_type for elem in elems}
        # Replace block types with templated
        for typ in list(types):
            if typ.startswith("block_"):  # pragma: no cover
                types.remove(typ)
                types.add("templated")
        # Take the easy route if they're all the same type
        if len(types) == 1:
            return types.pop()
        # Then deal with priority
        priority = ["templated", "escaped", "literal"]
        for p in priority:
            if p in types:
                return p
        raise RuntimeError(
            f"Exhausted priorities in _coalesce_types! {types!r}"
        )  # pragma: no cover

    @classmethod
    def _split_uniques_coalesce_rest(
        cls,
        split_file: list[IntermediateFileSlice],
        raw_occurrences: dict[str, list[int]],
        templ_occurrences: dict[str, list[int]],
        templated_str: str,
    ) -> Iterator[TemplatedFileSlice]:
        """Within each of the compound sections split on unique literals.

        For everything else we coalesce to the dominant type.

        Returns:
            Iterable of the type of segment, the slice in the raw file
                and the slice in the templated file.

        """
        # A buffer to capture tail segments
        tail_buffer: list[TemplatedFileSlice] = []

        templater_logger.debug("    _split_uniques_coalesce_rest: %s", split_file)

        for int_file_slice in split_file:
            # Yield anything from the tail buffer
            if tail_buffer:  # pragma: no cover
                templater_logger.debug(
                    "        Yielding Tail Buffer [start]: %s", tail_buffer
                )
                yield from tail_buffer
                tail_buffer = []

            # Check whether we're handling a zero length slice.
            if (
                int_file_slice.templated_slice.stop
                - int_file_slice.templated_slice.start
                == 0
            ):  # pragma: no cover
                point_combo = int_file_slice.coalesce()
                templater_logger.debug(
                    "        Yielding Point Combination: %s", point_combo
                )
                yield point_combo
                continue

            # Yield anything simple
            try:
                simple_elem = int_file_slice.try_simple()
                templater_logger.debug("        Yielding Simple: %s", simple_elem)
                yield simple_elem
                continue
            except ValueError:
                pass

            # Trim ends and overwrite the current working copy.
            head_buffer, int_file_slice, tail_buffer = int_file_slice.trim_ends(
                templated_str=templated_str
            )
            if head_buffer:
                yield from head_buffer  # pragma: no cover
            # Have we consumed the whole thing?
            if not int_file_slice.slice_buffer:
                continue  # pragma: no cover

            # Try to yield simply again (post trim)
            try:  # pragma: no cover
                simple_elem = int_file_slice.try_simple()
                templater_logger.debug("        Yielding Simple: %s", simple_elem)
                yield simple_elem
                continue
            except ValueError:
                pass

            templater_logger.debug("        Intermediate Slice: %s", int_file_slice)
            # Generate the coalesced version in case we need it
            coalesced = int_file_slice.coalesce()

            # Look for anchors
            raw_occs = cls._filter_occurrences(
                int_file_slice.source_slice, raw_occurrences
            )
            templ_occs = cls._filter_occurrences(
                int_file_slice.templated_slice, templ_occurrences
            )
            # Do we have any uniques to split on?
            # NB: We use `get` on the templated occurrences, because it's possible
            # that because of an if statement, something is in the source, but
            # not in the templated at all. In that case, we shouldn't use it.
            one_way_uniques = [
                key
                for key in raw_occs.keys()
                if len(raw_occs[key]) == 1 and len(templ_occs.get(key, [])) >= 1
            ]
            two_way_uniques = [
                key for key in one_way_uniques if len(templ_occs[key]) == 1
            ]
            # if we don't have anything to anchor on, then just return (coalescing
            # types)
            if not raw_occs or not templ_occs or not one_way_uniques:
                templater_logger.debug(
                    "        No Anchors or Uniques. Yielding Whole: %s", coalesced
                )
                yield coalesced
                continue

            # Deal with the inner segment itself.
            templater_logger.debug(
                "        Intermediate Slice [post trim]: %s: %r",
                int_file_slice,
                templated_str[int_file_slice.templated_slice],
            )
            templater_logger.debug("        One Way Uniques: %s", one_way_uniques)
            templater_logger.debug("        Two Way Uniques: %s", two_way_uniques)

            # Hang onto the starting position, which we'll advance as we go.
            starts = (
                int_file_slice.source_slice.start,
                int_file_slice.templated_slice.start,
            )

            # Deal with two way uniques first, because they are easier.
            # If we do find any we use recursion, because we'll want to do
            # all of the above checks again.
            if two_way_uniques:
                # Yield the uniques and coalesce anything between.
                bookmark_idx = 0
                for idx, raw_slice in enumerate(int_file_slice.slice_buffer):
                    pos = 0
                    unq: Optional[str] = None
                    # Does this element contain one of our uniques? If so, where?
                    for unique in two_way_uniques:
                        if unique in raw_slice.raw:
                            pos = raw_slice.raw.index(unique)
                            unq = unique

                    if unq:
                        # Yes it does. Handle it.

                        # Get the position of the unique section.
                        unique_position = (
                            raw_occs[unq][0],
                            templ_occs[unq][0],
                        )
                        templater_logger.debug(
                            "            Handling Unique: %r, %s, %s, %r",
                            unq,
                            pos,
                            unique_position,
                            raw_slice,
                        )

                        # Handle full slices up to this one
                        if idx > bookmark_idx:
                            # Recurse to deal with any loops separately
                            yield from cls._split_uniques_coalesce_rest(
                                [
                                    IntermediateFileSlice(
                                        "compound",
                                        # slice up to this unique
                                        slice(starts[0], unique_position[0] - pos),
                                        slice(starts[1], unique_position[1] - pos),
                                        int_file_slice.slice_buffer[bookmark_idx:idx],
                                    )
                                ],
                                raw_occs,
                                templ_occs,
                                templated_str,
                            )

                        # Handle any potential partial slice if we're part way through
                        # this one.
                        if pos > 0:
                            yield TemplatedFileSlice(
                                raw_slice.slice_type,
                                slice(unique_position[0] - pos, unique_position[0]),
                                slice(unique_position[1] - pos, unique_position[1]),
                            )

                        # Handle the unique itself and update the bookmark
                        starts = (
                            unique_position[0] + len(unq),
                            unique_position[1] + len(unq),
                        )
                        yield TemplatedFileSlice(
                            raw_slice.slice_type,
                            slice(unique_position[0], starts[0]),
                            slice(unique_position[1], starts[1]),
                        )
                        # Move the bookmark after this position
                        bookmark_idx = idx + 1

                        # Handle any remnant after the unique.
                        if raw_slice.raw[pos + len(unq) :]:
                            remnant_length = len(raw_slice.raw) - (len(unq) + pos)
                            _starts = starts
                            starts = (
                                starts[0] + remnant_length,
                                starts[1] + remnant_length,
                            )
                            yield TemplatedFileSlice(
                                raw_slice.slice_type,
                                slice(_starts[0], starts[0]),
                                slice(_starts[1], starts[1]),
                            )

                if bookmark_idx == 0:  # pragma: no cover
                    # This is a SAFETY VALVE. In Theory we should never be here
                    # and if we are it implies an error elsewhere. This clause
                    # should stop any potential infinite recursion in its tracks
                    # by simply classifying the whole of the current block as
                    # templated and just stopping here.
                    # Bugs triggering this eventuality have been observed in 0.4.0.
                    templater_logger.info(
                        "        Safety Value Info: %s, %r",
                        two_way_uniques,
                        templated_str[int_file_slice.templated_slice],
                    )
                    templater_logger.warning(
                        "        Python templater safety value unexpectedly triggered. "
                        "Please report your raw and compiled query on github for "
                        "debugging."
                    )
                    # NOTE: If a bug is reported here, this will incorrectly
                    # classify more of the query as "templated" than it should.
                    yield coalesced
                    continue

                # At the end of the loop deal with any remaining slices.
                # The above "Safety Valve"TM should keep us safe from infinite
                # recursion.
                if len(int_file_slice.slice_buffer) > bookmark_idx:
                    # Recurse to deal with any loops separately
                    yield from cls._split_uniques_coalesce_rest(
                        [
                            IntermediateFileSlice(
                                "compound",
                                # Slicing is easy here, we have no choice
                                slice(starts[0], int_file_slice.source_slice.stop),
                                slice(starts[1], int_file_slice.templated_slice.stop),
                                # Calculate the subsection to deal with.
                                int_file_slice.slice_buffer[
                                    bookmark_idx : len(int_file_slice.slice_buffer)
                                ],
                            )
                        ],
                        raw_occs,
                        templ_occs,
                        templated_str,
                    )
                # We continue here because the buffer should be exhausted,
                # and if there's more to do we'll do it in the recursion.
                continue

            # If we get here, then there ARE uniques, but they are only ONE WAY.
            # This means loops. Loops are tricky.
            # We're very unlikely to get here (impossible?) with just python
            # formatting, but this class is also the base for the jinja templater
            # (and others?) so it may be used there.
            # One way uniques give us landmarks to try and estimate what to do with
            # them.
            owu_templ_tuples = cls._sorted_occurrence_tuples(  # pragma: no cover
                {key: templ_occs[key] for key in one_way_uniques}
            )

            templater_logger.debug(  # pragma: no cover
                "        Handling One Way Uniques: %s", owu_templ_tuples
            )

            # Hang onto out *ending* position too from here.
            stops = (  # pragma: no cover
                int_file_slice.source_slice.stop,
                int_file_slice.templated_slice.stop,
            )

            # OWU in this context refers to "One Way Unique"
            this_owu_idx: Optional[int] = None  # pragma: no cover
            last_owu_idx: Optional[int] = None  # pragma: no cover
            # Iterate through occurrence tuples of the one-way uniques.
            for raw, template_idx in owu_templ_tuples:  # pragma: no cover
                raw_idx = raw_occs[raw][0]
                raw_len = len(raw)

                # Find the index of this owu in the slice_buffer, store the previous
                last_owu_idx = this_owu_idx
                try:
                    this_owu_idx = next(
                        idx
                        for idx, slc in enumerate(int_file_slice.slice_buffer)
                        if slc.raw == raw
                    )
                except StopIteration:  # pragma: no cover
                    # This can happen if the unique was detected, but was introduced
                    # by a templater step. This is a false positive. Skip and move on.
                    templater_logger.info(
                        "One Way Unique %r not found in slice buffer. Skipping...", raw
                    )
                    continue

                templater_logger.debug(
                    "        Handling OWU: %r @%s (raw @%s) [this_owu_idx: %s, "
                    "last_owu_dx: %s]",
                    raw,
                    template_idx,
                    raw_idx,
                    this_owu_idx,
                    last_owu_idx,
                )

                if template_idx > starts[1]:
                    # Yield the bit before this literal. We yield it
                    # all as a tuple, because if we could do any better
                    # we would have done it by now.

                    # Can we identify a meaningful portion of the patch
                    # to recurse a split?
                    sub_section: Optional[list[RawFileSlice]] = None
                    # If it's the start, the slicing is easy
                    if (
                        starts[1] == int_file_slice.templated_slice.stop
                    ):  # pragma: no cover TODO?
                        sub_section = int_file_slice.slice_buffer[:this_owu_idx]
                    # If we are AFTER the previous in the template, then it's
                    # also easy. [assuming it's not the same owu]
                    elif (
                        raw_idx > starts[0] and last_owu_idx != this_owu_idx
                    ):  # pragma: no cover
                        if last_owu_idx:
                            sub_section = int_file_slice.slice_buffer[
                                last_owu_idx + 1 : this_owu_idx
                            ]
                        else:
                            sub_section = int_file_slice.slice_buffer[:this_owu_idx]

                    # If we succeeded in one of the above, we can also recurse
                    # and be more intelligent with the other sections.
                    if sub_section:
                        templater_logger.debug(
                            "        Attempting Subsplit [pre]: %s, %r",
                            sub_section,
                            templated_str[slice(starts[1], template_idx)],
                        )
                        yield from cls._split_uniques_coalesce_rest(
                            [
                                IntermediateFileSlice(
                                    "compound",
                                    # Slicing is easy here, we have no choice
                                    slice(starts[0], raw_idx),
                                    slice(starts[1], template_idx),
                                    sub_section,
                                )
                            ],
                            raw_occs,
                            templ_occs,
                            templated_str,
                        )
                    # Otherwise, it's the tricky case.
                    else:
                        # In this case we've found a literal, coming AFTER another
                        # in the templated version, but BEFORE (or the same) in the
                        # raw version. This only happens during loops, but it means
                        # that identifying exactly what the intervening bit refers
                        # to is a bit arbitrary. In this case we're going to OVER
                        # estimate and refer to the whole loop segment.

                        # TODO: Maybe this should make two chunks instead, one
                        # working backward, and one working forward. But that's
                        # a job for another day.

                        # First find where we are starting this remainder
                        # in the template (as an index in the buffer).
                        # Any segments *after* cur_idx are involved.
                        if last_owu_idx is None or last_owu_idx + 1 >= len(
                            int_file_slice.slice_buffer
                        ):
                            cur_idx = 0
                        else:
                            cur_idx = last_owu_idx + 1

                        # We need to know how many block_ends are after this.
                        block_ends = sum(
                            slc.slice_type == "block_end"
                            for slc in int_file_slice.slice_buffer[cur_idx:]
                        )
                        # We can allow up to this number of preceding block starts
                        block_start_indices = [
                            idx
                            for idx, slc in enumerate(
                                int_file_slice.slice_buffer[:cur_idx]
                            )
                            if slc.slice_type == "block_start"
                        ]

                        # Trim anything which we're not allowed to use.
                        if len(block_start_indices) > block_ends:  # pragma: no cover
                            offset = block_start_indices[-1 - block_ends] + 1
                            elem_sub_buffer = int_file_slice.slice_buffer[offset:]
                            cur_idx -= offset
                        else:
                            elem_sub_buffer = int_file_slice.slice_buffer

                        # We also need to know whether any of the *starting*
                        # segments are involved.
                        # Anything up to start_idx (exclusive) is included.
                        include_start = raw_idx > elem_sub_buffer[0].source_idx

                        # The ending point of this slice, is already decided.
                        end_point = elem_sub_buffer[-1].end_source_idx()

                        # If start_idx is None, we're in luck. We don't need to include
                        # the beginning.
                        if include_start:
                            start_point = elem_sub_buffer[0].source_idx
                        # Otherwise we know it's looped round, we need to include the
                        # whole slice.
                        else:  # pragma: no cover
                            start_point = elem_sub_buffer[cur_idx].source_idx

                        tricky = TemplatedFileSlice(
                            "templated",
                            slice(start_point, end_point),
                            slice(starts[1], template_idx),
                        )

                        templater_logger.debug(
                            "        Yielding Tricky Case : %s",
                            tricky,
                        )

                        yield tricky

                # Yield the literal
                owu_literal_slice = TemplatedFileSlice(
                    "literal",
                    offset_slice(raw_idx, raw_len),
                    offset_slice(template_idx, raw_len),
                )
                templater_logger.debug(
                    "    Yielding Unique: %r, %s",
                    raw,
                    owu_literal_slice,
                )
                yield owu_literal_slice
                # Update our bookmark
                starts = (
                    raw_idx + raw_len,
                    template_idx + raw_len,
                )

            if starts[1] < stops[1] and last_owu_idx is not None:  # pragma: no cover
                # Yield the end bit
                templater_logger.debug("        Attempting Subsplit [post].")
                yield from cls._split_uniques_coalesce_rest(
                    [
                        IntermediateFileSlice(
                            "compound",
                            # Slicing is easy here, we have no choice
                            slice(raw_idx + raw_len, stops[0]),
                            slice(starts[1], stops[1]),
                            int_file_slice.slice_buffer[last_owu_idx + 1 :],
                        )
                    ],
                    raw_occs,
                    templ_occs,
                    templated_str,
                )

        # Yield anything from the tail buffer
        if tail_buffer:  # pragma: no cover
            templater_logger.debug(
                "        Yielding Tail Buffer [end]: %s", tail_buffer
            )
            yield from tail_buffer



"""Templater Code."""

from collections.abc import Iterator

# Although these shouldn't usually be instantiated from here
# we import them to make sure they get registered.
from sqlfluff.core.templaters.base import RawFileSlice, RawTemplater, TemplatedFile
from sqlfluff.core.templaters.jinja import JinjaTemplater
from sqlfluff.core.templaters.placeholder import PlaceholderTemplater
from sqlfluff.core.templaters.python import PythonTemplater


def core_templaters() -> Iterator[type[RawTemplater]]:
    """Returns the templater tuples for the core templaters."""
    yield from [
        RawTemplater,
        JinjaTemplater,
        PythonTemplater,
        PlaceholderTemplater,
    ]


__all__ = (
    "RawFileSlice",
    "TemplatedFile",
    "RawTemplater",
    "JinjaTemplater",
    "PythonTemplater",
    "PlaceholderTemplater",
    "core_templaters",
)



"""The Apache Doris dialect.

This dialect extends MySQL grammar with specific Apache Doris syntax features.
"""

from sqlfluff.core.dialects import load_raw_dialect
from sqlfluff.core.parser import (
    BaseSegment,
    Bracketed,
    CodeSegment,
    Delimited,
    Matchable,
    MultiStringParser,
    OneOf,
    OptionallyBracketed,
    Ref,
    SegmentGenerator,
    Sequence,
)
from sqlfluff.dialects import dialect_mysql as mysql
from sqlfluff.dialects.dialect_doris_keywords import (
    doris_reserved_keywords,
    doris_unreserved_keywords,
)

mysql_dialect = load_raw_dialect("mysql")
doris_dialect = mysql_dialect.copy_as(
    "doris",
    formatted_name="Apache Doris",
    docstring="""**Default Casing**: ``lowercase``

    **Quotes**: String Literals: ``''``, ``"``,
    Identifiers: |back_quotes|.

    The dialect for `Apache Doris <https://doris.apache.org/>`_.""",
)
doris_dialect.update_keywords_set_from_multiline_string(
    "unreserved_keywords", doris_unreserved_keywords
)
doris_dialect.sets("reserved_keywords").clear()
doris_dialect.update_keywords_set_from_multiline_string(
    "reserved_keywords", doris_reserved_keywords
)


# Add the engine types set for Doris
doris_dialect.sets("engine_types").update(
    ["olap", "mysql", "elasticsearch", "hive", "hudi", "iceberg", "jdbc", "broker"]
)


doris_dialect.add(
    EngineTypeSegment=SegmentGenerator(
        lambda dialect: MultiStringParser(
            dialect.sets("engine_types"),
            CodeSegment,
            type="engine_type",
        )
    ),
)


class ColumnDefinitionSegment(mysql.ColumnDefinitionSegment):
    """A column definition, e.g. for CREATE TABLE or ALTER TABLE.

    Doris-specific version that supports aggregation functions like
    MAX, MIN, REPLACE, SUM.
    """

    match_grammar = mysql.ColumnDefinitionSegment.match_grammar.copy(
        insert=[
            OneOf(
                "MAX",
                "MIN",
                "REPLACE",
                "SUM",
                "BITMAP_UNION",
                "HLL_UNION",
                "QUANTILE_UNION",
                optional=True,
            ),
        ]
    )


class CreateTableStatementSegment(mysql.CreateTableStatementSegment):
    """A `CREATE TABLE` statement.

    Doris-specific version that handles:
    - Different ENGINE types
    - Key types (DUPLICATE, AGGREGATE, UNIQUE)
    - Cluster by clause
    - Specific partition syntax
    - Distribution syntax
    - Rollup definitions
    - Specific table properties
    - CREATE TABLE ... AS SELECT (CTAS)
    - CREATE TABLE ... LIKE
    """

    type = "create_table_statement"
    match_grammar = Sequence(
        "CREATE",
        Ref("OrReplaceGrammar", optional=True),
        Ref.keyword("TEMPORARY", optional=True),
        Ref.keyword("EXTERNAL", optional=True),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        OneOf(
            # Standard column definitions
            Sequence(
                Bracketed(
                    Delimited(
                        OneOf(
                            Ref("TableConstraintSegment"),
                            Ref("ColumnDefinitionSegment"),
                            Ref("IndexDefinitionSegment"),
                        ),
                    )
                ),
                # Doris specific
                Sequence(
                    "ENGINE",
                    Ref("EqualsSegment"),
                    Ref("EngineTypeSegment"),
                    optional=True,
                ),
                # Key type
                Sequence(
                    OneOf(
                        Sequence("DUPLICATE", "KEY"),
                        Sequence("AGGREGATE", "KEY"),
                        Sequence("UNIQUE", "KEY"),
                    ),
                    Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                    # Cluster by clause
                    Sequence(
                        "CLUSTER",
                        "BY",
                        Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                        optional=True,
                    ),
                    optional=True,
                ),
                Ref("CommentClauseSegment", optional=True),
                # Partitioning
                Ref("PartitionSegment", optional=True),
                # Distribution
                Ref("DistributionSegment", optional=True),
                # Rollup definitions
                Ref("RollupSegment", optional=True),
                # Properties
                Sequence(
                    "PROPERTIES",
                    Bracketed(
                        Delimited(
                            Sequence(
                                Ref("QuotedLiteralSegment"),
                                Ref("EqualsSegment"),
                                Ref("QuotedLiteralSegment"),
                            )
                        )
                    ),
                    optional=True,
                ),
            ),
            # Create table like
            Sequence("LIKE", Ref("TableReferenceSegment")),
            # Create table as select (CTAS)
            Sequence(
                # Optional ENGINE clause
                Sequence(
                    "ENGINE",
                    Ref("EqualsSegment"),
                    Ref("EngineTypeSegment"),
                    optional=True,
                ),
                # Optional COMMENT clause
                Ref("CommentClauseSegment", optional=True),
                # Optional properties before AS SELECT
                Sequence(
                    "PROPERTIES",
                    Bracketed(
                        Delimited(
                            Sequence(
                                Ref("QuotedLiteralSegment"),
                                Ref("EqualsSegment"),
                                Ref("QuotedLiteralSegment"),
                            )
                        )
                    ),
                    optional=True,
                ),
                "AS",
                OptionallyBracketed(Ref("SelectableGrammar")),
            ),
        ),
    )


class ColumnConstraintSegment(mysql.ColumnConstraintSegment):
    """A column option; each CREATE TABLE column can have 0 or more."""

    match_grammar: Matchable = OneOf(
        mysql.ColumnConstraintSegment.match_grammar,
        Sequence("AS", Ref("ExpressionSegment")),
        Sequence("GENERATED", "ALWAYS", "AS", Bracketed(Ref("ExpressionSegment"))),
    )


class PartitionSegment(BaseSegment):
    """A partition segment supporting Doris specific syntax.

    Supports:
    1. Auto partitioning (AUTO PARTITION BY RANGE)
    2. Manual range partitioning (PARTITION BY RANGE)
    3. List partitioning (PARTITION BY LIST)
    """

    type = "partition_segment"
    match_grammar = OneOf(
        # Auto partitioning
        Sequence(
            "AUTO",
            "PARTITION",
            "BY",
            "RANGE",
            Bracketed(Ref("FunctionSegment")),
            Bracketed(),  # Empty partition list for auto partitioning
        ),
        # Manual partitioning
        Sequence(
            "PARTITION",
            "BY",
            OneOf(
                # Manual range partitioning
                Sequence(
                    "RANGE",
                    Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                    Bracketed(
                        Delimited(
                            OneOf(
                                Ref("RangePartitionDefinitionSegment"),
                                Ref("RangePartitionIntervalSegment"),
                            )
                        )
                    ),
                ),
                # List partitioning
                Sequence(
                    "LIST",
                    Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                    Bracketed(Delimited(Ref("ListPartitionDefinitionSegment"))),
                ),
            ),
        ),
    )


class RangePartitionDefinitionSegment(BaseSegment):
    """Range partition definition with VALUES LESS THAN or VALUES range."""

    type = "range_partition_definition"
    match_grammar = Sequence(
        "PARTITION",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("ObjectReferenceSegment"),
        "VALUES",
        OneOf(
            Sequence(
                "LESS",
                "THAN",
                OneOf(
                    "MAXVALUE",
                    Bracketed(Delimited(Ref("LiteralGrammar"))),
                ),
            ),
            Sequence(
                Bracketed(
                    Bracketed(Delimited(Ref("LiteralGrammar"))),
                    ",",
                    Bracketed(Delimited(Ref("LiteralGrammar"))),
                )
            ),
        ),
        # Partition properties
        Sequence(
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("QuotedLiteralSegment"),
                        Ref("EqualsSegment"),
                        Ref("QuotedLiteralSegment"),
                    )
                )
            ),
            optional=True,
        ),
    )


class RangePartitionIntervalSegment(BaseSegment):
    """Range partition definition with FROM TO INTERVAL syntax."""

    type = "range_partition_interval"
    match_grammar = Sequence(
        "FROM",
        Bracketed(Ref("QuotedLiteralSegment")),
        "TO",
        Bracketed(Ref("QuotedLiteralSegment")),
        "INTERVAL",
        Ref("NumericLiteralSegment"),
        OneOf("YEAR", "MONTH", "DAY", "HOUR", "MINUTE", "SECOND"),
    )


class ListPartitionDefinitionSegment(BaseSegment):
    """List partition definition with VALUES IN syntax."""

    type = "list_partition_definition"
    match_grammar = Sequence(
        "PARTITION",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("ObjectReferenceSegment"),
        "VALUES",
        "IN",
        Bracketed(
            Delimited(
                OneOf(
                    Bracketed(Delimited(Ref("LiteralGrammar"))),
                    Ref("LiteralGrammar"),
                )
            )
        ),
    )


class DistributionSegment(BaseSegment):
    """A distribution segment supporting both hash and random distribution."""

    type = "distribution_segment"
    match_grammar = Sequence(
        "DISTRIBUTED",
        "BY",
        OneOf(
            Sequence(
                "HASH",
                Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                Sequence(
                    "BUCKETS",
                    OneOf(
                        Ref("NumericLiteralSegment"),
                        "AUTO",
                    ),
                    optional=True,
                ),
            ),
            Sequence(
                "RANDOM",
                Sequence(
                    "BUCKETS",
                    OneOf(
                        Ref("NumericLiteralSegment"),
                        "AUTO",
                    ),
                    optional=True,
                ),
            ),
        ),
    )


class RollupSegment(BaseSegment):
    """Rollup definition for Doris tables."""

    type = "rollup_segment"
    match_grammar = Sequence(
        "ROLLUP",
        Bracketed(
            Delimited(
                Sequence(
                    Ref("ObjectReferenceSegment"),
                    Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                    Sequence(
                        "DUPLICATE",
                        "KEY",
                        Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                        optional=True,
                    ),
                )
            )
        ),
    )


class IndexDefinitionSegment(BaseSegment):
    """Index definition specific to Doris."""

    type = "index_definition"
    match_grammar = Sequence(
        "INDEX",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("IndexReferenceSegment"),
        Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
        Sequence("USING", OneOf("INVERTED", "BITMAP", "BLOOM_FILTER"), optional=True),
        Sequence(
            "PROPERTIES",
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("QuotedLiteralSegment"),
                        Ref("EqualsSegment"),
                        Ref("QuotedLiteralSegment"),
                    )
                )
            ),
            optional=True,
        ),
        Sequence("COMMENT", Ref("QuotedLiteralSegment"), optional=True),
    )


class DropTableStatementSegment(BaseSegment):
    """A `DROP TABLE` statement.

    Doris-specific version that supports:
    - IF EXISTS clause
    - Database-qualified table names
    - FORCE option
    """

    type = "drop_table_statement"
    match_grammar = Sequence(
        "DROP",
        "TABLE",
        Ref("IfExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),  # Single table reference, not delimited
        Sequence("FORCE", optional=True),  # Optional FORCE keyword
    )


class InsertStatementSegment(BaseSegment):
    """A `INSERT` statement.

    Doris-specific version that supports:
    - PARTITION clause
    - WITH LABEL clause
    - DEFAULT values
    """

    type = "insert_statement"
    match_grammar = Sequence(
        "INSERT",
        Ref.keyword("INTO", optional=True),
        Ref("TableReferenceSegment"),
        # Optional PARTITION clause
        Sequence(
            "PARTITION",
            Bracketed(Delimited(Ref("SingleIdentifierGrammar"))),
            optional=True,
        ),
        # Optional WITH LABEL clause
        Sequence(
            "WITH",
            "LABEL",
            Ref("SingleIdentifierGrammar"),
            optional=True,
        ),
        # Optional column list
        Ref("BracketedColumnReferenceListGrammar", optional=True),
        # VALUES or SELECT
        OneOf(
            Ref("ValuesClauseSegment"),
            Ref("SelectableGrammar"),
        ),
    )



"""Implementation of Rule CP01."""

from typing import Any, Callable, Optional

import regex

from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.config_info import get_config_info
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler

try:
    import sqlfluffrs

    _HAS_SQLFLUFFRS = True
except ImportError:  # pragma: no cover
    _HAS_SQLFLUFFRS = False


def is_capitalizable(character: str) -> bool:
    """Does the character have differing lower and upper-case versions?"""
    if character.lower() == character.upper():
        return False
    return True


class Rule_CP01(BaseRule):
    """Inconsistent capitalisation of keywords.

    **Anti-pattern**

    In this example, ``select`` is in lower-case whereas ``FROM`` is in upper-case.

    .. code-block:: sql

        select
            a
        FROM foo

    **Best practice**

    Make all keywords either in upper-case or in lower-case.

    .. code-block:: sql

        SELECT
            a
        FROM foo

        -- Also good

        select
            a
        from foo
    """

    name = "capitalisation.keywords"
    aliases = ("L010",)
    groups: tuple[str, ...] = ("all", "core", "capitalisation")
    is_fix_compatible = True

    lint_phase = "post"
    # Binary operators behave like keywords too.
    crawl_behaviour = SegmentSeekerCrawler({"keyword", "binary_operator", "date_part"})
    # Skip literals (which are also keywords) as they have their own rule (CP04)
    _exclude_types: tuple[str, ...] = ("literal",)
    _exclude_parent_types: tuple[str, ...] = (
        "data_type",
        "datetime_type_identifier",
        "primitive_type",
    )
    config_keywords = ["capitalisation_policy", "ignore_words", "ignore_words_regex"]
    # Human readable target elem for description
    _description_elem = "Keywords"

    def _eval_rust(self, context: RuleContext) -> Optional[list[LintResult]]:
        """Rust-native CP01 detection over the arena.

        Runs the whole detection in Rust (`sqlfluffrs.cp01_violations` over
        `root._rs_tree`) and maps each `(leaf_index, fixed_raw)` back to its
        Python segment to emit a standard `LintFix`. Returns ``None`` (Python
        fallback) when the Rust extension/arena is unavailable, or for any
        config the native path doesn't implement (a non-keyword policy or a
        regex word-ignore). See ``BaseRule._eval_rust``.
        """
        if not _HAS_SQLFLUFFRS:
            return None  # pragma: no cover
        # Only CP01 itself: this detection targets keyword/operator segments, so
        # subclasses without their own _eval_rust (currently CP02/CP05) must not
        # use it — they fall back to Python until they get their own Rust path.
        # CP04 in particular shares `capitalisation_policy`, so a policy check
        # alone wouldn't exclude it (it has its own override, see CP04.py).
        if self.name != "capitalisation.keywords":
            return None
        if getattr(self, "ignore_words_regex", None):
            return None
        # CP01's capitalisation_policy is validated to one of these four values.
        policy = str(getattr(self, "capitalisation_policy"))
        return self._eval_rust_capitalisation(
            context, sqlfluffrs.cp01_violations, policy
        )

    def _eval_rust_capitalisation(
        self,
        context: RuleContext,
        violations_fn: Callable[[Any, str, list[str], bool], list[tuple[int, str]]],
        policy: str,
    ) -> Optional[list[LintResult]]:
        """Shared Rust-dispatch plumbing for the capitalisation bundle.

        Each rule with a Rust path (CP01/CP03/CP04) does its own guard checks
        in its own ``_eval_rust`` (rule identity, arena presence via `None`
        propagation, `ignore_words_regex`), then delegates here for the FFI
        call and leaf-index -> segment mapping shared by all of them, keyed on
        which native `*_violations` function and policy config key the rule
        uses.
        """
        root = context.segment
        rs_tree = getattr(root, "_rs_tree", None)
        if rs_tree is None:
            # Unreachable via dispatch (crawl() only calls this when the root
            # has an arena), but guard in case _eval_rust is called directly.
            return None  # pragma: no cover

        ignore_words_config = str(getattr(self, "ignore_words", None))
        if ignore_words_config and ignore_words_config != "None":
            ignore_words = self.split_comma_separated_string(
                ignore_words_config.lower()
            )
        else:
            ignore_words = []
        ignore_templated = bool(context.config.get("ignore_templated_areas"))

        violations = violations_fn(rs_tree, policy, ignore_words, ignore_templated)
        if not violations:
            return []

        raw_segments = root.raw_segments
        policy_text = self._policy_description(policy)
        results: list[LintResult] = []
        for leaf_idx, fixed_raw in violations:
            segment = raw_segments[leaf_idx]
            results.append(
                LintResult(
                    anchor=segment,
                    fixes=[self._get_fix(segment, fixed_raw)],
                    description=f"{self._description_elem} must be {policy_text}",
                )
            )
        return results

    @staticmethod
    def _policy_description(policy: str) -> str:
        """Human-readable policy text for a Rust-path violation description."""
        if policy == "capitalise":
            return "capitalised."
        if policy == "consistent":
            return "consistent."
        return f"{policy} case."

    def _eval(self, context: RuleContext) -> Optional[list[LintResult]]:
        """Inconsistent capitalisation of keywords.

        We use the `memory` feature here to keep track of cases known to be
        INconsistent with what we've seen so far as well as the top choice
        for what the possible case is.

        """
        # NOTE: Given the dialect structure we can assume the targets have a parent.
        parent: BaseSegment = context.parent_stack[-1]
        if context.segment.is_type(*self._exclude_types) or parent.is_type(
            *self._exclude_parent_types
        ):
            return [LintResult(memory=context.memory)]

        # Used by CP03 (that inherits from this rule)
        # If it's a qualified function_name (i.e with more than one part to
        # function_name). Then it is likely an existing user defined function (UDF)
        # which are case sensitive so ignore for this.
        if parent.get_type() == "function_name" and len(parent.segments) != 1:
            return [LintResult(memory=context.memory)]

        return [self._handle_segment(context.segment, context)]

    def _handle_segment(self, segment: BaseSegment, context: RuleContext) -> LintResult:
        # NOTE: this mutates the memory field.
        memory = context.memory
        self.logger.info("_handle_segment: %s, %s", segment, segment.get_type())
        # Config type hints
        self.ignore_words_regex: str

        # Get the capitalisation policy configuration.
        try:
            cap_policy = self.cap_policy
            cap_policy_opts = self.cap_policy_opts
            ignore_words_list = self.ignore_words_list
            ignore_templated_areas = self.ignore_templated_areas
        except AttributeError:
            # First-time only, read the settings from configuration. This is
            # very slow.
            (
                cap_policy,
                cap_policy_opts,
                ignore_words_list,
                ignore_templated_areas,
            ) = self._init_capitalisation_policy(context)

        # Skip if in ignore list
        if ignore_words_list and segment.raw.lower() in ignore_words_list:
            return LintResult(memory=memory)

        # Skip if matches ignore regex
        if self.ignore_words_regex and regex.search(
            self.ignore_words_regex, segment.raw
        ):
            return LintResult(memory=memory)

        # Skip if templated.  If the user wants to ignore templated areas, we don't
        # even want to look at them to avoid affecting flagging non-template areas
        # that are inconsistent with the template areas.
        if segment.is_templated and ignore_templated_areas:
            return LintResult(memory=memory)

        # Skip if empty.
        if not segment.raw:
            return LintResult(memory=memory)

        refuted_cases = memory.get("refuted_cases", set())

        # Which cases are definitely inconsistent with the segment?
        first_letter_is_lowercase = False
        for character in segment.raw:
            if is_capitalizable(character):
                first_letter_is_lowercase = character != character.upper()
                break

        # We refute inference of camel, pascal, and snake case.
        # snake, if not explicitly set, can be destructive to
        # variable names, adding underscores.
        # camel and Pascal could allow poorly linted code in,
        # so must be explicitly chosen.
        refuted_cases.update(["camel", "pascal", "snake"])
        if first_letter_is_lowercase:
            refuted_cases.update(["upper", "capitalise"])
            if segment.raw != segment.raw.lower():
                refuted_cases.update(["lower"])
        else:
            refuted_cases.update(["lower"])
            if segment.raw != segment.raw.upper():
                refuted_cases.update(["upper"])
            if segment.raw != segment.raw.capitalize():
                refuted_cases.update(["capitalise"])

        # Update the memory
        memory["refuted_cases"] = refuted_cases

        self.logger.debug(
            f"Refuted cases after segment '{segment.raw}': {refuted_cases}"
        )

        # Skip if no inconsistencies, otherwise compute a concrete policy
        # to convert to.
        if cap_policy == "consistent":
            possible_cases = [c for c in cap_policy_opts if c not in refuted_cases]
            self.logger.debug(
                f"Possible cases after segment '{segment.raw}': {possible_cases}"
            )
            if possible_cases:
                # Save the latest possible case and skip
                memory["latest_possible_case"] = possible_cases[0]
                self.logger.debug(
                    f"Consistent capitalization, returning with memory: {memory}"
                )
                return LintResult(memory=memory)
            else:
                concrete_policy = memory.get("latest_possible_case", "upper")
                self.logger.debug(
                    f"Getting concrete policy '{concrete_policy}' from memory"
                )
        else:
            if cap_policy not in refuted_cases:
                # Skip
                self.logger.debug(
                    f"Consistent capitalization {cap_policy}, returning with "
                    f"memory: {memory}"
                )
                return LintResult(memory=memory)
            else:
                concrete_policy = cap_policy
                self.logger.debug(
                    f"Setting concrete policy '{concrete_policy}' from cap_policy"
                )

        # Set the fixed to same as initial in case any of below don't match
        fixed_raw = segment.raw
        # We need to change the segment to match the concrete policy
        if concrete_policy in ["upper", "lower", "capitalise"]:
            if concrete_policy == "upper":
                fixed_raw = fixed_raw.upper()
            elif concrete_policy == "lower":
                fixed_raw = fixed_raw.lower()
            elif concrete_policy == "capitalise":
                fixed_raw = fixed_raw.capitalize()
        elif concrete_policy == "pascal":
            # For Pascal we set the first letter in each "word" to uppercase
            # We do not lowercase other letters to allow for PascalCase style
            # words. This does mean we allow all UPPERCASE and also don't
            # correct Pascalcase to PascalCase, but there's only so much we can
            # do. We do correct underscore_words to Underscore_Words.
            fixed_raw = regex.sub(
                "([^a-zA-Z0-9]+|^)([a-zA-Z0-9])([a-zA-Z0-9]*)",
                lambda match: match.group(1) + match.group(2).upper() + match.group(3),
                segment.raw,
            )
        elif concrete_policy == "camel":
            # Similar to Pascal, for Camel, we can only do a best efforts approach.
            # This presents as us never changing case mid-string.
            fixed_raw = regex.sub(
                "([^a-zA-Z0-9]+|^)([a-zA-Z0-9])([a-zA-Z0-9]*)",
                lambda match: match.group(1) + match.group(2).lower() + match.group(3),
                segment.raw,
            )
        elif concrete_policy == "snake":
            if segment.raw.isupper():
                fixed_raw = segment.raw.lower()
            else:
                fixed_raw = regex.sub(
                    r"(?<=[a-z0-9])([A-Z])|(?<=[A-Za-z])([0-9])|(?<=[0-9])([A-Za-z])",
                    lambda match: "_" + match.group(),
                    segment.raw,
                ).lower()

        if fixed_raw == segment.raw:
            # No need to fix
            self.logger.debug(
                f"Capitalisation of segment '{segment.raw}' already OK with "
                f"policy '{concrete_policy}', returning with memory {memory}"
            )
            return LintResult(memory=memory)
        else:
            # build description based on the policy in use
            consistency = "consistently " if cap_policy == "consistent" else ""

            if concrete_policy in ["upper", "lower", "pascal", "camel", "snake"]:
                policy = f"{concrete_policy} case."
            elif concrete_policy == "capitalise":
                policy = "capitalised."

            # Return the fixed segment
            self.logger.debug(
                f"INCONSISTENT Capitalisation of segment '{segment.raw}', "
                f"fixing to '{fixed_raw}' and returning with memory {memory}"
            )
            return LintResult(
                anchor=segment,
                fixes=[self._get_fix(segment, fixed_raw)],
                memory=memory,
                description=f"{self._description_elem} must be {consistency}{policy}",
            )

    def _get_fix(self, segment: BaseSegment, fixed_raw: str) -> LintFix:
        """Given a segment found to have a fix, returns a LintFix for it.

        May be overridden by subclasses, which is useful when the parse tree
        structure varies from this simple base case.
        """
        return LintFix.replace(segment, [segment.edit(fixed_raw)])

    def _init_capitalisation_policy(self, context: RuleContext):
        """Called first time rule is evaluated to fetch & cache the policy."""
        cap_policy_name = next(
            k for k in self.config_keywords if k.endswith("capitalisation_policy")
        )
        self.cap_policy = getattr(self, cap_policy_name)
        valid_options = get_config_info()[cap_policy_name]["validation"] or []
        self.cap_policy_opts = [opt for opt in valid_options if opt != "consistent"]
        # Use str() as CP04 uses bools which might otherwise be read as bool
        ignore_words_config = str(getattr(self, "ignore_words"))
        if ignore_words_config and ignore_words_config != "None":
            self.ignore_words_list = self.split_comma_separated_string(
                ignore_words_config.lower()
            )
        else:
            self.ignore_words_list = []
        self.ignore_templated_areas = context.config.get("ignore_templated_areas")
        self.logger.debug(
            f"Selected '{cap_policy_name}': '{self.cap_policy}' from options "
            f"{self.cap_policy_opts}"
        )
        cap_policy = self.cap_policy
        cap_policy_opts = self.cap_policy_opts
        ignore_words_list = self.ignore_words_list
        ignore_templated_areas = self.ignore_templated_areas
        return cap_policy, cap_policy_opts, ignore_words_list, ignore_templated_areas



"""Implementation of Rule AM09."""

from typing import Optional, Tuple

from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_AM09(BaseRule):
    """Use of LIMIT and OFFSET without ORDER BY may lead to non-deterministic results.

    When using ``LIMIT`` or ``OFFSET``, it's generally recommended to include
    an ``ORDER BY`` clause to ensure deterministic results.

    **Anti-pattern**

    The following query has LIMIT and OFFSET without ORDER BY, which may return
    different results in successive executions.

    .. code-block:: sql

        SELECT *
        FROM foo
        LIMIT 10 OFFSET 5;

    **Best practice**

    Include an ``ORDER BY`` clause:

    .. code-block:: sql

        SELECT *
        FROM foo
        ORDER BY id
        LIMIT 10 OFFSET 5;
    """

    name = "ambiguous.order_by_limit"
    aliases = ()
    groups: Tuple[str, ...] = ("all", "ambiguous")
    crawl_behaviour = SegmentSeekerCrawler({"select_statement"})

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Check if LIMIT and OFFSET are used without ORDER BY."""
        segment = context.segment

        # Ensure it's a SELECT statement
        if not segment.is_type("select_statement"):
            return None

        # Detect presence of LIMIT and OFFSET
        has_limit = segment.get_child("limit_clause")
        has_offset = segment.get_child("offset_clause")
        has_order_by = segment.get_child("orderby_clause")

        # If LIMIT or OFFSET exist but ORDER BY is missing, issue a warning
        if (has_limit or has_offset) and not has_order_by:
            # Use the first relevant segment (LIMIT or OFFSET) as the anchor
            anchor_segment = has_limit or has_offset
            return LintResult(
                anchor=anchor_segment,
                description=(
                    "LIMIT and OFFSET are used without ORDER BY,"
                    " which may lead to non-deterministic results."
                ),
            )

        return None  # No issues found



"""Implementation of Rule AM08."""

from typing import Optional

from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_AM08(BaseRule):
    """Implicit cross join detected.

    **Anti-pattern**

    Cross joins are valid, but rare in the wild - and more often created by mistake
    than on purpose. This rule catches situations where a cross join has been specified,
    but not explicitly and so the risk of a mistaken cross join is highly likely.

    .. code-block:: sql
       :force:

        SELECT
            foo
        FROM bar
        JOIN baz;

    **Best practice**

    Use CROSS JOIN.

    .. code-block:: sql
       :force:

        SELECT
            foo
        FROM bar
        CROSS JOIN baz;
    """

    name = "ambiguous.join_condition"
    aliases = ()
    groups: tuple[str, ...] = ("all", "ambiguous")
    crawl_behaviour = SegmentSeekerCrawler({"join_clause"})
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Find joins without ON clause.

        Fix them into CROSS JOIN (if dialect allows it).
        """
        if not self._cross_join_supported(context):  # pragma: no cover
            # At the time of implementation, all dialects supports CROSS JOIN syntax.
            # Therefore, no cover is used on if statement.
            return None

        # We are only interested in JOIN clauses.
        join_clause = context.segment
        assert join_clause.is_type("join_clause")

        join_clause_keywords = [
            seg for seg in join_clause.segments if seg.type == "keyword"
        ]

        if any(
            kw.raw_upper in ("CROSS", "POSITIONAL", "USING", "NATURAL")
            for kw in join_clause_keywords
        ):
            # If explicit CROSS JOIN is used, disregard lack of condition
            # If explicit POSITIONAL JOIN is used, disregard lack of condition
            # If explicit NATURAL JOIN is used, disregard lack of condition
            # If explicit JOIN USING is used, disregard lack of condition
            return None

        this_join_condition = join_clause.get_child("join_on_condition")
        if this_join_condition:
            # Join condition is present, no error reported.
            return None

        select_stmt = self._get_select_stmt(context.parent_stack)
        if select_stmt is None:
            # Do not emit this warning for JOIN in UPDATE or DELETE
            return None

        maybe_where_clause = select_stmt.get_child("where_clause")
        if maybe_where_clause:
            # See CV12
            return None

        join_keywords = [kw for kw in join_clause_keywords if kw.raw_upper == "JOIN"]
        if len(join_keywords) != 1:
            # This can happen in T-SQL CROSS APPLY / OUTER APPLY
            return None

        # Skip if join is part of flattening logic
        maybe_from_expression_element = join_clause.get_child("from_expression_element")
        if maybe_from_expression_element:
            for (
                function_name_identifier
            ) in maybe_from_expression_element.recursive_crawl(
                "function_name_identifier"
            ):
                if function_name_identifier.raw_upper == "UNNEST":
                    return None

        return LintResult(join_clause)

    @staticmethod
    def _cross_join_supported(context: RuleContext) -> bool:
        return (
            False
            or "CROSS" in context.dialect.sets("reserved_keywords")
            or "CROSS" in context.dialect.sets("unreserved_keywords")
            or "CROSS" in context.dialect.sets("future_reserved_keywords")
        )

    @staticmethod
    def _get_select_stmt(stack: tuple[BaseSegment, ...]) -> Optional[BaseSegment]:
        for seg in reversed(stack):
            if seg.is_type("select_statement"):
                return seg
            elif seg.is_type("update_statement", "delete_statement"):
                return None

        # According to grammar, this is not reachable.
        # Do not emit any error instead of crashing.
        return None  # pragma: no cover



"""Process-global unique identifiers for segments.

Replaces per-instance ``uuid4()`` generation. A monotonic counter is
sufficient because segment ``uuid`` values are only ever used as opaque,
in-memory identity keys (dict keys and ``==`` in the fix engine). They never
need cryptographic unpredictability or cross-process uniqueness, and are never
parsed back as RFC-4122 UUIDs.

The high bits carry a source tag so these ids stay disjoint from token ids
minted on the Rust side (see the ``identity`` module in ``sqlfluffrs_types``,
which uses ``RUST_TAG = 2 << 120``). Both runtimes count from 1, so without the
tag a Python-created segment and a Rust-origin token would share low integer
ids -- and the fix engine would then treat two unrelated segments as the same
anchor. This is not hypothetical: ``PyToken.uuid`` surfaces the Rust ``u128``
to Python verbatim, so the two id spaces genuinely meet inside one tree.
"""

from __future__ import annotations

import itertools

# Source tag OR'd into the high bits of every Python-minted id. Must stay
# distinct from the Rust tag (``2 << 120``). Fits within 128 bits alongside any
# realistic process-lifetime counter value (which is always < ``2 ** 120``).
_PYTHON_TAG = 1 << 120

_counter = itertools.count(1)


def get_next_id() -> int:
    """Return a process-unique, source-tagged identifier.

    ``next()`` on an ``itertools.count`` is atomic under the CPython GIL -- its
    C-level ``__next__`` never releases the GIL mid-increment -- so this is
    thread-safe without an explicit lock. Avoiding the lock is the point: it
    keeps the per-segment cost to a single counter increment rather than
    re-introducing the lock acquire/release that motivated dropping ``uuid4``.
    """
    return _PYTHON_TAG | next(_counter)



"""A non-code matcher.

This is a stub of a grammar, intended for use entirely as a
terminator or similar alongside other matchers.
"""

from collections.abc import Sequence
from typing import Optional

from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.matchable import Matchable
from sqlfluff.core.parser.segments import BaseSegment
from sqlfluff.core.parser.types import SimpleHintType


class NonCodeMatcher(Matchable):
    """An object which behaves like a matcher to match non-code."""

    def simple(
        self, parse_context: ParseContext, crumbs: Optional[tuple[str, ...]] = None
    ) -> SimpleHintType:
        """This element doesn't work with simple."""
        return None

    def is_optional(self) -> bool:  # pragma: no cover
        """Not optional.

        NOTE: The NonCodeMatcher is only normally only used as a terminator
        or other special instance matcher. As such the `.simple()` method
        is unlikely to be used.
        """
        return False

    def cache_key(self) -> str:
        """Get the cache key for the matcher.

        NOTE: In this case, this class is a bit of a singleton
        and so we don't need a unique UUID in the same way as
        other classes.
        """
        return "non-code-matcher"

    def match(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> MatchResult:
        """Match any starting non-code segments."""
        matched_idx = idx
        for matched_idx in range(idx, len(segments)):
            if segments[matched_idx].is_code:
                break
        if matched_idx > idx:
            return MatchResult(matched_slice=slice(idx, matched_idx))
        # Otherwise return no match
        return MatchResult.empty_at(idx)



"""The postgres rules plugin bundle.

This plugin bundles linting rules which apply exclusively to PostgreSQL. At some
point in the future it might be useful to spin this off into a separate
installable python package, but so long as the number of rules remain
low, it makes sense to keep it bundled with SQLFluff core.
"""

from sqlfluff.core.plugin import hookimpl
from sqlfluff.core.rules import BaseRule


@hookimpl
def get_rules() -> list[type[BaseRule]]:
    """Get plugin rules.

    NOTE: Rules are imported only on fetch to manage import times
    when rules aren't used.
    """
    from sqlfluff.rules.postgres.PG01 import Rule_PG01
    from sqlfluff.rules.postgres.PG02 import Rule_PG02

    return [Rule_PG01, Rule_PG02]



"""A lookbehind matcher for keyword disambiguation.

This is a stub of a grammar, intended for use as an ``exclude`` pattern
on keyword terminators to prevent them from matching when preceded by
specific keyword sequences.
"""

from collections.abc import Sequence
from typing import Optional

from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.matchable import Matchable
from sqlfluff.core.parser.segments import BaseSegment
from sqlfluff.core.parser.types import SimpleHintType


class PrecededByMatcher(Matchable):
    """Matches when the current position is preceded by a keyword sequence.

    This is used as an ``exclude`` pattern on keyword matchers to prevent
    them from matching when preceded by specific keyword sequences.

    For example, it prevents ``FROM`` in ``IS [NOT] DISTINCT FROM`` from
    being treated as the start of a ``FROM`` clause when used as an exclude
    on the ``FROM`` keyword in ``SelectClauseTerminatorGrammar``.

    Args:
        preceding_sequences: Ordered tuples of uppercase keyword strings
            that may appear immediately before the current position, where
            matching is performed from right to left within each sequence
            (i.e. the last item is checked first). Non-code and meta
            segments between keywords are skipped.
    """

    def __init__(
        self,
        preceding_sequences: tuple[tuple[str, ...], ...],
    ) -> None:
        self.preceding_sequences = preceding_sequences

    def is_optional(self) -> bool:  # pragma: no cover
        """Return whether this element is optional.

        A lookbehind matcher is never optional — it must always be evaluated.
        """
        return False

    def simple(
        self, parse_context: ParseContext, crumbs: Optional[tuple[str, ...]] = None
    ) -> SimpleHintType:  # pragma: no cover
        """This element doesn't work with simple."""
        return None

    def cache_key(self) -> str:  # pragma: no cover
        """Get the cache key for the matcher."""
        return "preceded-by-" + "|".join(
            "-".join(sequence) for sequence in self.preceding_sequences
        )

    @staticmethod
    def _skip_non_code_and_meta_backward(
        segments: Sequence["BaseSegment"],
        idx: int,
    ) -> int:
        """Move ``idx`` backward past any non-code or meta segments."""
        while idx >= 0 and (not segments[idx].is_code or segments[idx].is_meta):
            idx -= 1
        return idx

    def match(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> MatchResult:
        """Match when the position is preceded by the configured keywords.

        Scans backward from ``idx`` through non-code and meta segments,
        checking each candidate keyword sequence defined in
        ``preceding_sequences``. Each sequence is checked right-to-left,
        i.e. the last element of a sequence must appear closest to the
        current position.

        Returns a non-empty :class:`MatchResult` if the lookbehind matches
        (meaning the caller's ``exclude`` should suppress the outer match),
        or an empty result if it does not match.
        """
        for preceding in self.preceding_sequences:
            if self._match_preceding_sequence(segments, idx, preceding):
                return MatchResult(slice(idx, idx + 1))

        return MatchResult.empty_at(idx)

    def _match_preceding_sequence(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        preceding: tuple[str, ...],
    ) -> bool:
        """Return whether a specific preceding sequence matches."""
        prev = idx - 1

        for keyword in reversed(preceding):
            prev = self._skip_non_code_and_meta_backward(segments, prev)
            if prev < 0 or segments[prev].raw_upper != keyword:
                return False
            prev -= 1

        return True


# Shared exclude pattern for the FROM keyword in select clause terminators.
# Prevents FROM in "IS [NOT] DISTINCT FROM" being treated as a FROM clause.
is_distinct_from_lookbehind = PrecededByMatcher(
    preceding_sequences=(("IS", "DISTINCT"), ("IS", "NOT", "DISTINCT")),
)



"""Implementation of Rule ST10."""

from collections.abc import Iterator

from sqlfluff.core.rules import BaseRule, EvalResultType, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_ST10(BaseRule):
    """Redundant constant expression.

    Including an expression that always evaluates to
    either ``TRUE`` or ``FALSE`` regardless of the input columns is
    unnecessary and makes statements harder to read and understand.

    Constant conditions are sometimes mistakes (by mistyping the column
    name intended), and sometimes the result of incorrect information that
    they are necessary in some circumstances. In the former case, they can
    sometimes result in a cartesian join if it was supposed to be a join
    condition. Given the ambiguity of intent, this rule does not suggest
    an automatic fix, and instead invites the user to resolve the problem
    manually.

    **Anti-pattern**

    .. code-block:: sql

        SELECT *
        FROM my_table
        -- This following WHERE clause is redundant.
        WHERE my_table.col = my_table.col

    **Best practice**

    .. code-block:: sql

        SELECT *
        FROM my_table
        -- Replace with a condition that includes meaningful logic,
        -- or remove the condition entirely.
        WHERE my_table.col > 3
    """

    name = "structure.constant_expression"
    aliases = ()
    groups: tuple[str, ...] = ("all", "structure")
    config_keywords = []
    crawl_behaviour = SegmentSeekerCrawler({"expression"})
    is_fix_compatible = False

    def _eval(self, context: RuleContext) -> EvalResultType:
        return [lint_result for lint_result in self._eval_gen(context)]

    def _eval_gen(self, context: RuleContext) -> Iterator[LintResult]:
        assert context.segment.is_type("expression")
        subsegments = context.segment.segments
        count_subsegments = len(subsegments)
        # The following literal expressions are allowable because they're
        # often included in auto-generated code.
        # NOTE: In future this could become a configuration option.
        allowable_literal_expressions = {"1 = 1", "1 = 0"}

        for idx, seg in enumerate(context.segment.segments):
            if seg.is_type("comparison_operator"):
                if seg.raw not in ("=", "!=", "<>"):
                    continue

                lhs_idx, lhs = next(
                    (
                        (i, subsegments[i])
                        for i in range(idx - 1, -1, -1)
                        if not subsegments[i].is_whitespace
                    ),
                    (None, None),
                )

                rhs_idx, rhs = next(
                    (
                        (i, subsegments[i])
                        for i in range(idx + 1, count_subsegments, 1)
                        if not subsegments[i].is_whitespace
                    ),
                    (None, None),
                )
                if lhs is None or rhs is None or lhs_idx is None or rhs_idx is None:
                    # Should be unreachable with correctly parsed tree
                    continue  # pragma: no cover

                if lhs.is_templated or rhs.is_templated:
                    continue

                prev_before_lhs = (
                    next(
                        (
                            subsegments[i]
                            for i in range(lhs_idx - 1, -1, -1)
                            if not subsegments[i].is_whitespace
                        ),
                        None,
                    )
                    if lhs_idx > 0
                    else None
                )

                next_after_rhs = next(
                    (
                        subsegments[i]
                        for i in range(rhs_idx + 1, count_subsegments, 1)
                        if not subsegments[i].is_whitespace
                    ),
                    None,
                )

                # We treat the pieces immediately left and right of `=` as its operands.
                # After AND or OR, that is correct (e.g. `... OR 1 = 2`).
                # After other binary operators, it is not correct (e.g. `num % 2 = 0`
                # on the left, or `flags = flags & @mask` on the right).
                if (
                    prev_before_lhs is not None
                    and prev_before_lhs.is_type("binary_operator")
                    and prev_before_lhs.raw_normalized().upper() not in ("AND", "OR")
                ):
                    continue
                if (
                    next_after_rhs is not None
                    and next_after_rhs.is_type("binary_operator")
                    and next_after_rhs.raw_normalized().upper() not in ("AND", "OR")
                ):
                    continue

                # literals need explicit handling (due to well-defined allow-list)
                if lhs.is_type("literal") and rhs.is_type("literal"):
                    expr_s = f"{lhs.raw_normalized()} {seg.raw} {rhs.raw_normalized()}"
                    if expr_s in allowable_literal_expressions:
                        # ignore based on allowlist
                        continue
                else:
                    if lhs.type != rhs.type:
                        continue
                    if lhs.raw_normalized() != rhs.raw_normalized():
                        continue

                # attach violation to eq/ne operator in expression
                yield LintResult(seg)
